from flax import struct
import jax
import jax.numpy as jnp

@struct.dataclass
class Signal:
    content:dict

@struct.dataclass
class Params:
    content:dict

@struct.dataclass
class State:
    content:dict

@struct.dataclass
class Policy:
    state:State
    params:Params
    deterministic:bool

    @staticmethod
    def create_policy(params:Params, active_state:bool, key:jax.random.PRNGKey):
        pass
    @staticmethod
    def step_policy(input:Signal, policy:struct.dataclass):
        pass
    @staticmethod
    def reset_policy(input:Signal, policy:struct.dataclass):
        pass

@struct.dataclass
class Agent:
    params:Params
    state:State
    unique_id:jnp.int32
    agent_type:jnp.int32
    active_state:bool
    age:jnp.float32
    policy: Policy

    @staticmethod
    def create_agent(create_params:Params, unique_id:int, active_state:int, agent_type:int, key: jax.random.PRNGKey):
        pass
    
    @staticmethod
    def step_agent(step_params:Params, input:Signal, agents:struct.dataclass):
        pass

    @staticmethod
    def add_agent(add_params:Params, agents:struct.dataclass, idx:int, key:jax.random.PRNGKey):
        pass

    @staticmethod
    def set_agent(set_params:Params, agents:struct.dataclass, idx:int, key:jax.random.PRNGKey):
        pass

    @staticmethod
    def reset_agent():
        pass

    @staticmethod
    def remove_agent(remove_params:Params, Agents:struct.dataclass, idx:int):
        pass

class Agent_Set:
    agents:Agent
    def __init__(self, agent:Agent, num_total_agents:jnp.int32, num_active_agents:jnp.int32, agent_type:jnp.int32):

        self.create_agents = jax.vmap(agent.create_agent, in_axes=(None,0,0,0,0))
        #create( params, unique_id, active_state, agent_types, key)
        #params is not vmaped over

        self.step_agents = jax.jit(jax.vmap(agent.step_agent, in_axes=(None,0,0)))

        self.reset_agents = jax.jit(jax.vmap(agent.reset_agent))

        self.num_total_agents = num_total_agents
        self.num_active_agents = num_active_agents
        self.num_inactive_agents = num_total_agents - num_active_agents
        self.agent_type = agent_type
        print("AgentSet initialized")




#this policy class is used just to transfer the keys and show the structure of the policy class


from flax.struct import dataclass
from base.agent_classes import Policy, Signal, Params, State
import jax
import jax.numpy as jnp

class Random_Policy(Policy):

    @staticmethod
    def create_policy(params:Params, key:jax.random.PRNGKey):
        state = State(content={'key':key})
        return Policy(state=state, params=params, deterministic=False)
    
    @staticmethod
    def step_policy(input:Signal, policy:Policy):
        key = policy.state.content['key']
        key, subkey = jax.random.split(key)
        action = jax.random.uniform(key, shape=(1,))
        state = State(content={'key':subkey})
        new_policy = policy.replace(state=state)
        return Signal(content={'action':action}), new_policy
    
    @staticmethod
    def reset_policy(key:jax.random.PRNGKey, policy:Policy):
        state = State(content={'key':key})
        return policy.replace(state=state)



from flax import struct 
import jax
import jax.numpy as jnp




@struct.dataclass
class Point:
    x: jnp.float32
    y: jnp.float32
    def create_point(x:float, y:float):
        return Point(x, y)

@struct.dataclass
class Wall:
    p1: Point
    p2: Point
    def create_wall(p1:Point, p2:Point):
        return Wall(p1, p2)

@struct.dataclass
class Ray:
    ray_origin: Point
    ray_direction: Point # cosines and sines of the angle
    ray_length: jnp.float32
    def create_ray(ray_origin:Point, ray_direction:Point, ray_length:float):
        return Ray(ray_origin, ray_direction, ray_length)



@struct.dataclass
class Space:
    x_min: jnp.float32
    x_max: jnp.float32
    y_min: jnp.float32
    y_max: jnp.float32
    torous: bool
    walls: Wall
    def create_space(x_min, x_max, y_min, y_max, torous, wall_begins:Point, wall_ends:Point):
        x_min = x_min
        x_max = x_max
        y_min = y_min
        y_max = y_max
        torous = torous
        walls = jax.vmap(Wall.create_wall)(wall_begins, wall_ends)
        return Space(x_min, x_max, y_min, y_max, torous, walls)
    




@struct.dataclass
class NetworkSpace:
    pass



from base.space_classes import *
import jax
import jax.numpy as jnp





def check_collision(walls: Wall, old_position:Point, new_position:Point):
    def check_one_wall(wall : Wall, old_position:Point, new_position:Point):

        def on_segment(p:Point, q:Point, r:Point):
            cond = jnp.logical_and(q.x <= jnp.maximum(p.x, r.x), 
                                   jnp.logical_and(q.x >= jnp.minimum(p.x, r.x), 
                                    jnp.logical_and(q.y <= jnp.maximum(p.y, r.y), q.y >= jnp.minimum(p.y, r.y))))
            return cond
    
        def orientation(p:Point, q:Point, r:Point):
            val = (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y)
            return jax.lax.select(val == 0, 0, jax.lax.select(val > 0, 1, 2))
        
        def do_intersect(wall : Wall, old_position:Point, new_position:Point):
            p1 = wall.p1
            q1 = wall.p2
            p2 = old_position
            q2 = new_position
            o1 = orientation(p1, q1, p2)
            o2 = orientation(p1, q1, q2)
            o3 = orientation(p2, q2, p1)
            o4 = orientation(p2, q2, q1)
            
            c1 = jnp.logical_and(o1 != o2, o3 != o4)
            c2 = jnp.logical_and(o1 == 0, on_segment(p1, p2, q1))
            c3 = jnp.logical_and(o2 == 0, on_segment(p1, q2, q1))
            c4 = jnp.logical_and(o3 == 0, on_segment(p2, p1, q2))
            c5 = jnp.logical_and(o4 == 0, on_segment(p2, q1, q2))
            return jnp.logical_or(jnp.logical_or(c1, c2), jnp.logical_or(c3, jnp.logical_or(c4, c5)))
        
        return do_intersect(wall, old_position, new_position)
    
    return jnp.any(jax.vmap(check_one_wall, in_axes=(0, None, None))(walls, old_position, new_position)) # any rertuns true if any wall intersects with the particle
jit_check_collision = jax.jit(check_collision)

def wall_generator(wall_array:jnp.array):# wall_array = [[[wall1_begin(x,y)], [wall1_end(x,y)]], [[wall2_begin(x,y)], [wall2_end(x,y)], ...]]])
    x_wall_begins = jnp.array([wall_array[i][0][0] for i in range(len(wall_array))])
    y_wall_begins = jnp.array([wall_array[i][0][1] for i in range(len(wall_array))])
    x_wall_ends = jnp.array([wall_array[i][1][0] for i in range(len(wall_array))])
    y_wall_ends = jnp.array([wall_array[i][1][1] for i in range(len(wall_array))])
    wall_points_begins = jax.vmap(Point.create_point)(x_wall_begins, y_wall_begins)
    wall_points_ends = jax.vmap(Point.create_point)(x_wall_ends, y_wall_ends)
    #walls = jax.vmap(Wall.create_wall)(wall_points_begins, wall_points_ends)
    return wall_points_begins, wall_points_ends

RESOLUTION = 9
def ray_generator(agent_pos, ray_span, ray_length):
    x,y,angle = agent_pos
    ray_angles = jnp.linspace(angle - ray_span, angle + ray_span, RESOLUTION)

    cos_ray_angles = jnp.cos(ray_angles)
    sin_ray_angles = jnp.sin(ray_angles)
    ray_directions = jax.vmap(Point.create_point)(cos_ray_angles, sin_ray_angles)

    ray_origin = Point.create_point(x, y)
    
    rays = jax.vmap(Ray.create_ray, in_axes=(None, 0, None))(ray_origin, ray_directions, ray_length)

    return rays
jit_ray_generator = jax.jit(ray_generator)

def ray_agent_sensor(ray, data):
    agent_x, agent_y, agent_rad, agent_active_status = data
    def ray_active_agent_sensing(ray, agent_x, agent_y, agent_rad):
        ray_origin = jnp.array([ray.ray_origin.x, ray.ray_origin.y])
        ray_origin = jnp.reshape(ray_origin, (2,))
        ray_direction = jnp.array([ray.ray_direction.x, ray.ray_direction.y])
        ray_direction = jnp.reshape(ray_direction, (2,))
        obj_center = jnp.array([agent_x, agent_y])
        obj_radius = agent_rad
        s = ray_origin - obj_center
        b = jnp.dot(s, ray_direction)
        c = jnp.dot(s, s) - obj_radius**2
        h = b**2 - c
        h = jax.lax.cond(h < 0, lambda _: -1.0, lambda _: jnp.sqrt(h), None)
        t = jax.lax.cond(h >= 0, lambda _: -b - h, lambda _: ray.ray_length, None)
        t = jax.lax.cond(t < 0, lambda _: ray.ray_length, lambda _: t, None)
        return jnp.minimum(t, ray.ray_length)

        
    def ray_inactive_agent_sensing():
        return ray.ray_length
    distance = jax.lax.cond(agent_active_status, lambda _: ray_active_agent_sensing(ray, agent_x, agent_y, agent_rad), lambda _: ray_inactive_agent_sensing(), None)
    return distance
jit_ray_agent_sensor = jax.jit(ray_agent_sensor)

def ray_wall_sensor(ray, data):
    wall_p1_x, wall_p1_y, wall_p2_x, wall_p2_y = data
    def line_intercept(p1:jnp.array, q1:jnp.array, p2:jnp.array, q2:jnp.array):
        d1 = jnp.cross(q1 - p1, p2 - p1)
        d2 = jnp.cross(q1 - p1, q2 - p1)
        r = d1 / ((d1 - d2)+1e-6)
        intersection = p2 + r * (q2 - p2)
        return jnp.linalg.norm(intersection - p1)
        
    def on_segment(p:jnp.array, q:jnp.array, r:jnp.array):
        cond = jnp.logical_and(q[0] <= jnp.maximum(p[0], r[0]), 
                                jnp.logical_and(q[0] >= jnp.minimum(p[0], r[0]), 
                                jnp.logical_and(q[1] <= jnp.maximum(p[1], r[1]), q[1] >= jnp.minimum(p[1], r[1])))
                            )
        return cond
    def orientation(p:jnp.array, q:jnp.array, r:jnp.array):
        val = (q[1] - p[1]) * (r[0] - q[0]) - (q[0] - p[0]) * (r[1] - q[1])
        return jax.lax.select(val == 0, 0, jax.lax.select(val > 0, 1, 2))
        
    def do_intersect(ray:Ray, wall_p1_x, wall_p1_y, wall_p2_x, wall_p2_y):
        p1 = jnp.array([ray.ray_origin.x, ray.ray_origin.y])
        p1 = jnp.reshape(p1, (2,))
        q1 = jnp.array([ray.ray_origin.x + ray.ray_direction.x * ray.ray_length, ray.ray_origin.y + ray.ray_direction.y * ray.ray_length])
        q1 = jnp.reshape(q1, (2,))
        p2 = jnp.array([wall_p1_x, wall_p1_y])
        q2 = jnp.array([wall_p2_x, wall_p2_y])
        o1 = orientation(p1, q1, p2)
        o2 = orientation(p1, q1, q2)
        o3 = orientation(p2, q2, p1)
        o4 = orientation(p2, q2, q1)
            
        c1 = jnp.logical_and(o1 != o2, o3 != o4)
        d1 = jax.lax.cond(c1, lambda _: line_intercept(p1, q1, p2, q2), lambda _: ray.ray_length, None)
            
        c2 = jnp.logical_and(o1 == 0, on_segment(p1, p2, q1))
        d2 = jax.lax.cond(c2, lambda _: jnp.linalg.norm(p1 - p2), lambda _: ray.ray_length, None)
            
        c3 = jnp.logical_and(o2 == 0, on_segment(p1, q2, q1))
        d3 = jax.lax.cond(c3, lambda _: jnp.linalg.norm(p1 - q2), lambda _: ray.ray_length, None)
            
        c4 = jnp.logical_and(o3 == 0, on_segment(p2, p1, q2))
        d4 = jax.lax.cond(c4, lambda _: 0.0, lambda _: ray.ray_length, None)

        c5 = jnp.logical_and(o4 == 0, on_segment(p2, q1, q2))
        d5 = jax.lax.cond(c5, lambda _: jnp.linalg.norm(p1 - q1), lambda _: ray.ray_length, None)
        return jnp.minimum(d1, jnp.minimum(d2, jnp.minimum(d3, jnp.minimum(d4, d5))))
    return do_intersect(ray, wall_p1_x, wall_p1_y, wall_p2_x, wall_p2_y)
jit_ray_wall_sensor = jax.jit(ray_wall_sensor)


def create_space(x_min, x_max, y_min, y_max, torous, wall_array):
    x_min = x_min
    x_max = x_max
    y_min = y_min
    y_max = y_max
    torous = torous
    if wall_array is None:
        return Space(x_min, x_max, y_min, y_max, torous, None)
    wall_begin_points, wall_end_points = wall_generator(wall_array)
    walls = jax.vmap(Wall.create_wall)(wall_begin_points, wall_end_points)
    return Space(x_min, x_max, y_min, y_max, torous, walls)


import sys
sys.path.insert(1, '/home/siddhart/source/Foragax_v2')
from base.agent_classes import Policy, Signal, Params, State

import jax
import jax.numpy as jnp

class WilsonCowan(Policy):
    @staticmethod
    def create_policy(params:Params, key:jax.random.PRNGKey):
        num_neurons = params.content['num_neurons']
        num_obs = params.content['num_obs']
        num_actions = params.content['num_actions']
        deterministic = params.content['deterministic']

        Z = jnp.zeros((num_neurons,))
        state = State(content={'Z':Z})
       
        key, *init_keys = jax.random.split(key, 6)
        init_keys = jnp.array(init_keys)

        J = jax.random.uniform(init_keys[0], (num_neurons,num_neurons), minval=-1, maxval=1)
        tau = jax.random.uniform(init_keys[1], (num_neurons,), minval=-1, maxval=1)
        E = jax.random.uniform(init_keys[2], (num_neurons,num_obs), minval=-1, maxval=1)
        B = jax.random.uniform(init_keys[3], (num_neurons,), minval=-1, maxval=1)
        D = jax.random.uniform(init_keys[4], (num_actions,num_neurons), minval=-1, maxval=1)
        dt = params.content['dt']
        action_scale = params.content['action_scale']
        
        params = Params(content={'J':J, 'tau':tau, 'E':E, 'B':B, 'D':D, 'dt':dt, 'action_scale':action_scale})
        return Policy(state=state, params=params, deterministic=deterministic)

    @staticmethod
    def step_policy(input:Signal, policy:Policy):
        J = policy.params.content['J']
        tau = policy.params.content['tau']
        E = policy.params.content['E']
        B = policy.params.content['B']
        D = policy.params.content['D']
        dt = policy.params.content['dt']
        action_scale = policy.params.content['action_scale']

        Z = policy.state.content['Z']
        obs = input.content['obs']

        dz = jnp.matmul(J, jnp.tanh(Z)) + jnp.matmul(E, obs) + B - Z
        dz = jnp.multiply(dz, 10*jax.nn.sigmoid(tau))
        new_Z = Z + dt*dz # maybe remove dt to make the steps bigger
        new_policy = policy.replace(state=State(content={'Z':new_Z}))
        actions = action_scale * jnp.tanh(jnp.matmul(D, jnp.tanh(new_Z)))
        return Signal(content={'actions':actions}), new_policy

    @staticmethod
    def reset_policy(key:jax.random.PRNGKey, policy:Policy):
        key, *init_keys = jax.random.split(key, 6)
        init_keys = jnp.array(init_keys)

        J = jax.random.uniform(init_keys[0], policy.params.content['J'].shape, minval=-1, maxval=1)
        tau = jax.random.uniform(init_keys[1], policy.params.content['tau'].shape, minval=-1, maxval=1)
        E = jax.random.uniform(init_keys[2], policy.params.content['E'].shape, minval=-1, maxval=1)
        B = jax.random.uniform(init_keys[3], policy.params.content['B'].shape, minval=-1, maxval=1)
        D = jax.random.uniform(init_keys[4], policy.params.content['D'].shape, minval=-1, maxval=1)

        new_params = Params(content={'J':J, 'tau':tau, 'E':E, 'B':B, 'D':D, 'dt':policy.params.content['dt'], 'action_scale':policy.params.content['action_scale']})
        new_policy = policy.replace(params=new_params)
        return new_policy

        


import jax
import jax.numpy as jnp
from flax import struct
from base.agent_classes import *
from base.space_classes import *



def create_agents(params:Params, agent_set:Agent_Set, key:jax.random.PRNGKey):
    uniq_ids = jnp.arange(agent_set.num_total_agents)
    key, *create_keys = jax.random.split(key, agent_set.num_total_agents+1)
    create_keys = jnp.array(create_keys)

    active_states = jnp.concatenate((jnp.ones(agent_set.num_active_agents), jnp.zeros(agent_set.num_inactive_agents)))
    agent_types = jnp.tile(agent_set.agent_type, agent_set.num_total_agents)

    return agent_set.create_agents(params, uniq_ids, active_states, agent_types, create_keys)


def step_agents(params:Params, input:Signal, agent_set:Agent_Set):
    try:
        return agent_set.step_agents(params, input, agent_set.agents)
    except ValueError:
        print("Error in step_agents")


def add_agents(add_func:callable, num_agents_add:int, add_params:Params, agents:Agent, key:jax.random.PRNGKey):
    id_last_active = jnp.sum(agents.active_state, dtype=jnp.int32)

    def set_data(idx, agents__add_params__key):
        agents, add_params, key = agents__add_params__key
        new_agent,key = jax.jit(add_func)(add_params, agents, idx, key)
        new_agents = jax.tree_util.tree_map(lambda x,y:x.at[idx].set(y), agents, new_agent)
        return new_agents, add_params, key
        
    new_agents, add_params, key = jax.lax.fori_loop(id_last_active, id_last_active+num_agents_add, set_data, (agents, add_params, key))
    return new_agents, key

jit_add_agents = jax.jit(add_agents, static_argnums=(0,))


def remove_agents(remove_func:callable, num_agents_remove:int, remove_params:Params, agents:Agent):
    
    def remove_data(idx, agents__remove_params):
        agents, remove_params = agents__remove_params
        remove_ids = remove_params.content['remove_ids']
        new_agent = jax.jit(remove_func)(remove_params, agents, remove_ids[idx])
        new_agents = jax.tree_util.tree_map(lambda x,y:x.at[remove_ids[idx]].set(y), agents, new_agent)
        return new_agents, remove_params
    new_agents, remove_params = jax.lax.fori_loop(0, num_agents_remove, remove_data, (agents, remove_params))
    return new_agents

jit_remove_agents = jax.jit(remove_agents, static_argnums=(0,))

def set_agents(set_func:callable, num_agents_set:int, set_params:Params, agents:Agent, key:jax.random.PRNGKey):
    def set_data(idx, agents__set_params__key):
        agents, set_params, key = agents__set_params__key
        set_ids = set_params.content['set_ids']
        new_agent, key = jax.jit(set_func)(set_params, agents, set_ids[idx], key)
        new_agents = jax.tree_util.tree_map(lambda x,y:x.at[set_ids[idx]].set(y), agents, new_agent)
        return new_agents, set_params, key
    new_agents, set_params, key = jax.lax.fori_loop(0, num_agents_set, set_data, (agents, set_params, key))
    return new_agents, key
jit_set_agents = jax.jit(set_agents, static_argnums=(0,))

#returns a sorted list of indexes of agents (1 for ids that are selected, 0 for ids that are not selected) + the length of selected ids 
def select_agents(select_func:bool, select_params:Params, agents:Agent):
    selected_ids = jnp.where(jax.jit(select_func)(agents, select_params), 1.0, 0.0)
    selected_ids = jnp.reshape(selected_ids,(-1,))
    sort_selected_ids = jnp.argsort(-1*selected_ids)    
    selected_ids_len = jnp.sum(selected_ids, dtype=jnp.int32)
    return selected_ids_len, sort_selected_ids
jit_select_agents = jax.jit(select_agents, static_argnums=(0,))

def sort_agents(quantity:jnp.array, ascend:bool, agents:Agent):
    quantity = jnp.reshape(quantity, (-1,))
    sorted_ids = jax.lax.cond(ascend, lambda _: jnp.argsort(quantity), lambda _: jnp.argsort(-quantity), None)
    new_agents = jax.tree_util.tree_map(lambda x: jnp.take(x, sorted_ids, axis=0), agents)
    return new_agents, sorted_ids
jit_sort_agents = jax.jit(sort_agents)


